"""
FROWNS LICENSE

Copyright (c) 2001-2003, Brian Kelley
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met: 

    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer. 
    * Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following disclaimer
      in the documentation and/or other materials  provided with the
      distribution. 
    * Neither the name of Brian Kelley nor the names of frowns
      contributors may be used to endorse or promote products derived
      from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
# Build a simple Molecule object given the events from the Smiles
# tokenizer.

import string
import handler
import weakref
from ..mol import Atom, Bond, Molecule

# bondlookup is of the form
# textSymbol, bondsymbol, bondorder, bondtype, equiv class, stereo

STEREO_NONE = None
STEREO_UP = "UP"
STEREO_DOWN = "DOWN"

BONDLOOKUP = {'-': ('-', 1, 1, 1, STEREO_NONE),
              '=': ('=', 2, 2, 2, STEREO_NONE),
              '#': ('#', 3, 3, 3, STEREO_NONE),
              '\\': ('\\',1, 1, 1, STEREO_DOWN),
              '/': ('/' ,1, 1, 1, STEREO_UP),
              ':':(':', 1.5, 4, 4, STEREO_NONE),              
              }

def get_symbol_aromatic(text):
    if text[0] in "cnosp":
        return string.upper(text), 1
    return text, 0

def normalize_closure(text):
    if text[:1] == "%":
        return int(text[1:])
    return int(text)

implicit_bond = -123

class DummyVFGraph:
    def __init__(self):
        self.atoms = -1
    def InsertNode(self, node):
        self.atoms += 1
        return self.atoms
    def InsertEdge(self, index1, index2, bond):
        pass
    
class BuildMol(handler.TokenHandler):

    def begin(self):
        self.closures = {}

        self.atoms = []
        self.bonds = []
        self._atom = None
        self._prev_atoms = []

        # None occurs after a '.'
        # implicit_bond means implicit single bond
        self._pending_bond = None

    def end(self):
        if len(self._prev_atoms) >= 2:
            raise AssertionError("Missing ')'")
        if self._pending_bond not in [implicit_bond, None]:
            raise AssertionError("Missing an atom after the bond")
        if self.closures:
            raise AssertionError("Missing closures for %s" %
                                 (self.closures.keys(),))
        self.mol = Molecule(self.atoms, self.bonds)
    
    def add_token(self, field, pos, text):
        getattr(self, "do_" + field)(text)

    def add_atom(self, atom):
        atoms = self.atoms
        atom.index = len(atoms)
        atoms.append(atom)
        
        if self._pending_bond == implicit_bond:
            # Implicit single or aromatic bond
            self._pending_bond = Bond()

        if self._pending_bond is not None:
            bond = self._pending_bond
            prev_atom = self._prev_atoms[-1]
            bond.atoms[:] = [prev_atom, atom]
            ##self.mol.add_bond(bond, prev_atom, atom)
            bond.atoms = [prev_atom, atom]
            atom.bonds.append(bond)
            prev_atom.bonds.append(bond)
            atom.oatoms.append(prev_atom)
            prev_atom.oatoms.append(atom)
            self.bonds.append(bond)
            
        self._pending_bond = implicit_bond
        if not self._prev_atoms:
            self._prev_atoms.append(atom)
        else:
            self._prev_atoms[-1] = atom

        #self.mol.atoms.append(atom)
        
    def do_raw_atom(self, text):
        atom = Atom()
        symbol, atom.aromatic = get_symbol_aromatic(text)
        atom.set_symbol(symbol)
        self.add_atom(atom)

    def do_open_bracket(self, text):
        self._atom = Atom()
        self._atom.has_explicit_hcount = True

    def do_weight(self, text):
        self._atom.weight = int(text)

    def do_element(self, text):
        symbol, self._atom.aromatic = get_symbol_aromatic(text)
        self._atom.set_symbol(symbol)
        
    def do_chiral_count(self, text):
        #print "setting chirality", self._atom, int(text[1:])
        self._atom.chirality = int(text[1:])

    def do_chiral_named(self, text):
        self._atom.chiral_class = text[1:3]
        self._atom.chirality = int(text[3:])

    def do_chiral_symbols(self, text):
        self._atom.chiral_class = len(text)

    def do_hcount(self, text):
        if text == "H":
            self._atom.explicit_hcount = 1
        else:
            self._atom.explicit_hcount = int(text[1:])

    def do_positive_count(self, text):
        self._atom.charge = int(text[1:])

    def do_positive_symbols(self, text):
        self._atom.charge = len(text)

    def do_negative_count(self, text):
        self._atom.charge = -int(text[1:])

    def do_negative_symbols(self, text):
        self._atom.charge = -len(text)

    def do_close_bracket(self, text):
        self.add_atom(self._atom)
        self._atom = None

    def do_bond(self, text):
        assert self._pending_bond in (implicit_bond, None)
        symbol, bondorder, bondtype, equiv_class, stereo = BONDLOOKUP[text]
        # if the bond came in as aromatic (which it
        #  CAN'T!))
        if bondtype == 4:
            assert 0, "Bond's shouldn't come in as ':'"
            fixed = 0
        else:
            fixed = 1
        bond = Bond(text, bondorder, bondtype, fixed, stereo)
        bond.equiv_class = equiv_class
        self._pending_bond = bond

    def do_dot(self, text):
        assert self._pending_bond in (implicit_bond, None)
        self._pending_bond = None

    def do_closure(self, text):
        num = normalize_closure(text)
        if self.closures.has_key(num):
            prev_atom, bond = self.closures[num]
            del self.closures[num]

            assert self._pending_bond is not None, "Can't happen"
            
            if self._pending_bond is not implicit_bond and \
               bond is not implicit_bond and \
               self._pending_bond.symbol != "-":  # according to toolkit

                # need to verify they are compatible
                prev_symbol = bond.symbol
                symbol = self._pending_bond.symbol
                if (prev_symbol == symbol) or \
                   (prev_symbol == "/" and symbol == "\\") or \
                   (prev_symbol == "\\" and symbol == "/"):
                    pass
                else:
                    raise AssertionError("bond types don't match")
            elif bond is implicit_bond and self._pending_bond is not implicit_bond:
                # see if one of the bonds is not implicit and keep it
                bond = self._pending_bond
            elif bond is implicit_bond:
                # both are implicit so make a new one
                bond = Bond()

            bond._closure = 1
            atom = self._prev_atoms[-1]
            if prev_atom is atom:
                raise AssertionError("cannot close a ring with itself")
            bond.atoms[:] = [prev_atom, atom]
            prev_atom._closure = 1
            atom._closure = 1
            ##self.mol.add_bond(bond, prev_atom, atom)
            
            bond.atoms = [prev_atom, atom]
            atom.bonds.append(bond)
            prev_atom.bonds.append(bond)
            atom.oatoms.append(prev_atom)
            prev_atom.oatoms.append(atom)
            self.bonds.append(bond)

        else:
            self.closures[num] = (self._prev_atoms[-1], self._pending_bond)
        self._pending_bond = implicit_bond
            

    def do_open_branch(self, text):
        self._prev_atoms.append(self._prev_atoms[-1])
    
    def do_close_branch(self, text):
        self._prev_atoms.pop()



"""
FROWNS LICENSE

Copyright (c) 2001-2003, Brian Kelley
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met: 

    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer. 
    * Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following disclaimer
      in the documentation and/or other materials  provided with the
      distribution. 
    * Neither the name of Brian Kelley nor the names of frowns
      contributors may be used to endorse or promote products derived
      from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.



Bond class
Bonds

"""
from .idgen import defaultGenerator
import weakref

#DAYLIGHT CIS/TRANS support
DX_CHI_CIS=1
DX_CHI_TRANS=2
DX_CHI_NO_DBO=3

#Chirality Support
UP = "UP"
DOWN = "DOWN"
EITHER = "EITHER"

class Bond(object):
    __slots__ = ["adjunct", "atoms", "index", "rings", "_closure",
                 "dbo", "stereo", "parent", "handle",
                 "symbol", "aromatic", "bondorder", "equiv_class",
                 "bondtype", "fixed", "_traverseOrder", "_line"]
    
    def __init__(self,
                 symbol="-",
                 bondorder=1,
                 bondtype=1,
                 fixed=0,
                 stereo=None,                
                 generator=defaultGenerator):
        self.reset(symbol, bondorder, bondtype, fixed, stereo)

        self.adjunct = 0
        self.atoms = []

        self.index = -1
        self.rings = []
        self._closure = 0
        self.dbo = None
        self.stereo = None
        self.parent = None
        
        self.handle = id(self) #generator() # generate a new unique
                               #   # handle
           
    def setdbo(self, bond1, bond2, dboval):
        """Set the double bond orientation for bond1 and bond2
        based on this bond"""
        # this bond must be a double bond
        if self.bondtype != 2:
            raise FrownsError("To set double bond order, center bond must be double!")
        assert dboval in [DX_CHI_CIS, DX_CHI_TRANS, DX_CHI_NO_DBO], "bad dboval value"
        
        self.dbo.append(bond1, bond2, dboval)
        
            
    def reset(self, symbol, bondorder, bondtype, fixed, stereo):
        self.symbol = symbol
        self.bondorder = bondorder
        self.equiv_class = self.bondtype = bondtype
        self.fixed = fixed
        self.stereo = stereo
        # set the aromaticity correctly
        if self.bondtype == 4:
            self.aromatic = 1
        else:
            assert self.bondtype == self.bondorder
            self.aromatic = 0
        
    def set_symbol(self, symbol):
        """(symbol, bondorder) -> set the bondsymbol
        of the molecule"""
        raise "Deprecated"
        self.symbol, self.bondtype, bondorder, self.equiv_class = \
                     BONDLOOKUP[symbol]
        if self.bondtype == 4:
            self.aromatic = 1
        else:
            self.aromatic = 0


    def set_parent(self, parent):
        self.parent = weakref.ref(parent)

    def destroy(self):
        self.parent = None
        if self.atoms:
            a1, a2 = self.atoms
            a1.bonds.remove(self)
            a2.bonds.remove(self)
            a1.oatoms.remove(a2)
            a2.oatoms.remove(a1)
        self.rings = []
                
    def xatom(self, atom):
        """(atom)->return the atom at the other end of this bond
        or None if atom is not part of this bond"""
        handle = atom.handle
        
        if handle == self.atoms[0].handle:
            return self.atoms[1]
        elif handle == self.atoms[1].handle:
            return self.atoms[0]
        return None
        

    def __repr__(self):
        return "%s(%s)"%(self.__class__.__name__,
                         self.index)




"""
FROWNS LICENSE

Copyright (c) 2001-2003, Brian Kelley
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met: 

    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer. 
    * Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following disclaimer
      in the documentation and/or other materials  provided with the
      distribution. 
    * Neither the name of Brian Kelley nor the names of frowns
      contributors may be used to endorse or promote products derived
      from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

from .atypes import defaultAtomTypes
from .idgen import defaultGenerator
import chirality
from ..exceptions import PinkyError

#
# Things might change a little bit, but here are some
# implementation specific details
#
# chiral ordering is not defined yet...
#
# self.rings -> what rings is this atom in...
#               I don't know if I need yet
#
# self._findbonds -> lookup table for the findbond function
#                    one of the ring detectors uses this
#                    function frequently, so an optimization
#                    was in order.

CLOCKWISE = "@@"
ANTICLOCKWISE = "@"

class Atom(object):
    __slots__ = ["symbol", "hcount", "explicit_hcount", "imp_hcount",
                 "charge", "weight", "aromatic", "chiral_order", "adjunct",
                 "equiv_class", "symclass", "symorder", "index",
                 "rings", "bonds", "oatoms", "_closure", "chirality",
                 "chiral_class", "x", "y", "z", "parent", "handle",
                 "valences", "number", "mass", "negativity", "name",
                 "_chirality", "_line", "has_explicit_hcount"
                ]
    def __init__(self, generator=defaultGenerator):
        self.symbol = None
        self.hcount = 0
        self.explicit_hcount = 0
        self.has_explicit_hcount = False
        self.imp_hcount = 0
        self.charge = 0
        self.weight = 0
        self.aromatic = 0
        self.chiral_order = None
        self.adjunct = 0
        self.equiv_class  = -1
        self.symclass = -1
        self.symorder = -1
        self.index = -1
        
        self.rings = []
        self.bonds = []
        self.oatoms = []
        self._closure = 0
        self.chirality = None
        self._chirality = None
        self.chiral_class = None

        # coordinates for drawing and the like
        self.x = None
        self.y = None
        self.z = None

        self._line = ""
        
        self.parent = None
        self.handle = id(self)#generator()
        
    def set_symbol(self, symbol):
        self.symbol, self.number, self.name, self.mass, self.negativity, \
                     self.valences, self.equiv_class = defaultAtomTypes[symbol]

## XXX implicit_hcount is the responsibility of the modifing code
##  (actually, everything is...)  maybe I should try
## properties

    def findbond(self, otherAtom):
        handle = otherAtom.handle
        for atom, bond in zip(self.oatoms, self.bonds):
            if handle == atom.handle:
                return bond
        return None

    def sumBondOrders(self):
        result = 0
        for x in self.bonds:
            if x.bondtype == 4:
                # XXX FIX ME
                # this is a hack to fix bad conjugation
                # this will be fixed soon
                result += 1.5
            else:
                result += x.bondorder
        return result

    def destroy(self):
        self.rings = []
        self.bonds = []
        self.oatoms = []

    def chival(self, bonds):
        """compute the chiral value around an atom given a list of bonds"""
        # XXX I'm not sure how this works?
        order = [bond.xatom(self) for bond in bonds]
        return self._chirality(order)
    
    def setchival(self, bondorder, rotation):
        """compute chiral ordering of surrounding atoms"""
        rotation = [None, "@", "@@"][(rotation % 2)]
        # check to see if the bonds are attached
        if not bondorder: # use the default xatoms
            if len(self.oatoms) < 3 and self.explicit_hcount != 1:
                raise PinkyError("Need to have an explicit hydrogen when specifying "\
                                  "chirality with less than three bonds")
                

            self._chirality = chirality.T(self.oatoms,
                                            rotation)
            return
        if len(bondorder) != len(self.bonds):
            raise AtomError("The order of all bonds must be specified")
        
        for bond in bondorder:
            if bond not in self.bonds:
                raise AtomError("Specified bonds to assign chirality are not attatched to atom")

        order = [bond.xatom(self) for bond in bonds]
        self._chirality = chirality.T(order, rotation)
        
    def __repr__(self):
        return "%s(%s)"%(self.__class__.__name__,
                         self.index)




"""perception

Potentially perceptable people perceive partial properties.

A lot of computation chemistry involves the perception of various
chemical properties.  The term perception is used because some
chemists have different beliefs about what defines chemical
properties.

The list of properties that frowns perceives is the following:

Smallest Set of Smallest Rings
Aromaticity (which depends on the smallest set of smallest rings)

These routines are designed to be flexible, that is they can be
relaplaced by other routines easily.

Each of the derived perception routines has the following
signature:

molecule = perception(molecule)

This format is retained because in the future a copy of the molecule
with the appropriate annotations might be returned.

Currently for the sake of speed the molecule is modified (not
a copy of the molecule)  This may change in the future.
"""




"""
FROWNS LICENSE

Copyright (c) 2001-2003, Brian Kelley
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met: 

    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer. 
    * Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following disclaimer
      in the documentation and/or other materials  provided with the
      distribution. 
    * Neither the name of Brian Kelley nor the names of frowns
      contributors may be used to endorse or promote products derived
      from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.



Cycle(atoms) -> cycle object for a ring.

   cycle.atoms -> atoms around a ring
   cycle.bonds -> bonds aroung a ring

   cycle.rotate(atom) -> rotate the ring so that atom is in front
"""
class CycleError(Exception):
    pass

class Cycle:
    def __init__(self, atoms, bonds, aromatic=0):
        """(atoms)->create a cycle object
        assumes that the atoms are in traversal order around the ring.
        That is [a1, a2, a3] means that there is a bond in the cycle
        between a1 and a2 and a2 and a3 and a3 and a1"""
        self.atoms = atoms[:]
        self.bonds = bonds[:]
        self.aromatic = aromatic
        for atom in self.atoms:
            atom.rings.append(self)
            
        for bond in self.bonds:
            bond.rings.append(self)
        
    def __len__(self):
        return len(self.atoms)

    def rotate(self, atom):
        """(atom)->start the cycle at position atom, assumes
        that atom is in the cycle"""
        try:
            index = self.atoms.index(atom)
        except ValueError:
            raise CycleError("atom %s not in cycle"%(atom))

        self.atoms = self.atoms[index:] + self.atoms[:index]
        self.bonds = self.bonds[index:] + self.bonds[:index]

    def clone(self):
        return Cycle(self.atoms, self.bonds, self.aromatic)

    def set_aromatic(self):
        """set the cycle to be an aromatic ring"""
        #XXX FIX ME
        # this probably shouldn't be here
        for atom in self.atoms:
            atom.aromatic = 1
            
        for bond in self.bonds:
            bond.aromatic = 1
            bond.bondorder = 1.5
            bond.bondtype = 4
            bond.symbol = ":"
            bond.fixed = 1

        self.aromatic = 1



"""
FROWNS LICENSE

Copyright (c) 2001-2003, Brian Kelley
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met: 

    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer. 
    * Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following disclaimer
      in the documentation and/or other materials  provided with the
      distribution. 
    * Neither the name of Brian Kelley nor the names of frowns
      contributors may be used to endorse or promote products derived
      from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


computes equivalence classes for atoms

equiv_class = compute_equiv_class(atom)
"""

def compute_equiv_class(atom):
    """(atom)->Computes a unique integer for an atom"""    
    try:
        equiv_class = atom.number + \
                      1000*(atom.charge+10) + \
                      100000*(atom.hcount) + \
                      1000000*(atom.weight)
    except TypeError:
        raise ValueError, \
              "Can't compute number from atom.number %s atom.charge %s atom.hcount %s"\
              " atom.weight %s"%(atom.number, atom.charge, atom.hcount, atom.weight)
    return equiv_class



import os
from unittest import TestCase
from nose.tools import *
from pinky.smiles import smilin
from pinky.exceptions import PinkyError

class SmilesTestCase(TestCase):
    def setUp(self):
        self.path = os.path.dirname(os.path.abspath(__file__))

    def test_smilein(self):
        smiles_strings = []
        with open("{}/smiles.txt".format(self.path), 'rb') as fh:
            for line in fh:
                smiles_strings.append(line.strip())

        for smile in smiles_strings:
            mol = smilin(smile)
            out = mol.arbsmiles()
            can = mol.cansmiles()
            for bond in mol.bonds:
                print bond.symbol, bond.bondorder, bond.bondtype, bond.fixed
            for atom in mol.atoms:
                print atom, atom.sumBondOrders()
                
            print smile, out, can



"""
FROWNS LICENSE

Copyright (c) 2001-2003, Brian Kelley
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met: 

    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer. 
    * Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following disclaimer
      in the documentation and/or other materials  provided with the
      distribution. 
    * Neither the name of Brian Kelley nor the names of frowns
      contributors may be used to endorse or promote products derived
      from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

SmilesTokens

This is a list of smiles tokens that define the smiles language.
There are for types of tokens

 Atoms - C or [1000C]
 Bonds - single, double, sacrifice fly
 closures - things like C1CCC1 (the ones are closures)
 branches (Branch Start, Branch End)
          - these are ( and ) respectively

 Each Token Class below can convert the appropriate object into
 a string through overloading the __str__ mechanism.
"""
#
# Canonical labeling rules for daylight atoms
# From the Daylight Theory Manual
# Organic subset atoms may be written without brackets if the number of
# attached hydrogens conforms to the lowest normal valence consistent
# with explicit bonds. "Lowest normal valences" are B (3), C (4),
# N (3,5), O (2), P (3,5), S (2,4,6), and 1 for the halogens.
# Atoms in aromatic rings are specified by lower case letters, e.g.,
# aliphatic carbon is represented by the capital letter C, aromatic
# carbon by lower case c.
ORGANIC_SUBSET = ['B', 'C', 'N', 'O', 'P', 'S', 'F', 'Cl', 'Br', 'I']

class Atom:
    def __init__(self, atom, closureMappings, idGenerator):
        self.atom = atom
        self.closures = []
        self.closureMappings = closureMappings
        self.idGenerator = idGenerator

    def name_atom(self):
        atom = self.atom
        symbol = "%s"%(atom.symbol,)
        weight = atom.weight
        charge = atom.charge
        hcount = atom.hcount
        explicit_hcount = atom.explicit_hcount
        
        if atom.aromatic:
            out_symbol = symbol[0].lower() + symbol[1:]
        else:
            out_symbol = symbol

        # pyrole like nitrogens
        if atom.aromatic and symbol == "N" and charge == 0 and \
           weight == 0 and explicit_hcount == 1:
            # XXX Fix Me
            # There should only be one of these per five membered
            # aromatic ring
            return "[nH]"

        if symbol in ORGANIC_SUBSET and atom.valences and \
           not weight and not charge:
            sumOrders = atom.sumBondOrders()
            hcount = atom.hcount
            for valence in atom.valences:
                if sumOrders + hcount == valence:
                    return out_symbol#+"<%s>"%atom.handle

        if not weight: weight = ""
        
        if charge == -1: charge = "-"
        elif charge == 1: charge = "+"
        elif charge > 1: charge = "+%s"%charge
        else: charge = ""

        if hcount == 1: hcount = "H"
        elif hcount == 0: hcount = ""
        elif hcount > 1: hcount = "H%s"%hcount
        else:
            raise "Negative hcount!!!"

        return "[%s%s%s%s]"%(weight, out_symbol, hcount, charge)
                    
    def __str__(self):
        # easy for now
        atom = self.atom
        label = [self.name_atom()]
        
        idGenerator = self.idGenerator
        orderOfClosures = []
        closureMappings = self.closureMappings
        for id, bond in self.closures:
            if not closureMappings.has_key(id):
                closureMappings[id] = idGenerator.next()
            id = closureMappings[id]
            orderOfClosures.append((id,bond))
        orderOfClosures.sort()

        for id, bond in orderOfClosures:
            if bond: label.append(str(bond))
            if id > 9:
                label.append("%%%s"%id)
            else:
                label.append("%s"%id)
        return "".join(label)
#
# XXX FIX ME -> a lot of shared code here, it would be nicer
# if we can get rid of them
class IsomericAtom(Atom):
    def name_atom(self):
        atom = self.atom
        symbol = "%s"%(atom.symbol,)
        weight = atom.weight
        charge = atom.charge
        hcount = atom.hcount
        explicit_hcount = atom.explicit_hcount
        chirality = atom._chirality
        
        if atom.aromatic:
            out_symbol = symbol[0].lower() + symbol[1:]
        else:
            out_symbol = symbol

        # pyrole like nitrogens
        if atom.aromatic and symbol == "N" and charge == 0 and \
           weight == 0 and explicit_hcount == 1:
            # XXX Fix Me
            # There should only be one of these per five membered
            # aromatic ring
            return "[nH]"

        if symbol in ORGANIC_SUBSET and atom.valences and \
           not weight and not charge and not chirality:
            sumOrders = atom.sumBondOrders()
            hcount = atom.hcount
            for valence in atom.valences:
                if sumOrders + hcount == valence:
                    return out_symbol#+"<%s>"%atom.handle

        if not weight: weight = ""
        
        if charge == -1: charge = "-"
        elif charge == 1: charge = "+"
        elif charge > 1: charge = "+%s"%charge
        else: charge = ""

        if hcount == 1: hcount = "H"
        elif hcount == 0: hcount = ""
        elif hcount > 1: hcount = "H%s"%hcount
        else:
            raise "Negative hcount!!!"

        if chirality:
            bonds = [(bond._traverseOrder, bond) for bond in atom.bonds]
            bonds.sort()
            bonds = [bond[1].xatom(atom) for bond in bonds]
            chiralstr = chirality.getChirality(bonds)
        else:
            chiralstr = ""
        return "[%s%s%s%s%s]"%(weight, out_symbol, chiralstr, hcount, charge)
        return "[%s%s%s%s<%s>]"%(weight, out_symbol, hcount, charge, atom.handle)
    
class SmartsAtom(Atom):
    def name_atom(self):
        atom = self.atom
        symbol = atom.symbol

        if atom.aromatic:
            symbol = symbol.lower()
            
        if atom.symbol in ORGANIC_SUBSET:
            return symbol
        else:
            return "[%s]"%symbol
    

class Bond:
    def __init__(self, bond):
        self.bond = bond
        self.lookup = {
            1:"",
            2:"=",
            3:"#",
            4:"",
            5:"\\",
            6:"/"}
        
    def __str__(self):
        return self.lookup[self.bond.bondtype]    

class SmartsBond(Bond):
    def __init__(self, bond):
        self.bond = bond
        self.lookup = {
            1:"-",
            2:"=",
            3:"#",
            4:":",
            5:"\\",
            6:"/"}
        
class Branch:       
    def __str__(self):
        return "("

class BranchEnd:
    def __str__(self):
        return ")"
